/*****************************************************************************
 * capture.c
 *
 * bc-cat module unit test - read texture from webcam
 *
 * Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *   
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the
 *   distribution.
 *   
 *   Neither the name of Texas Instruments Incorporated nor the names of
 *   its contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
#include <stdio.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/videodev2.h>
#include <errno.h>
#include <img_defs.h>
#include <servicesext.h>
#include <bc_cat.h>


typedef struct buf_info {
    int index;
    size_t length;
    char *start;
} buf_info_t;

#define TRUE                1
#define FALSE               0

#define NUM_BUFFERS         3
#define CAPTURE_DEVICE      "/dev/video0"
#define CAPTURE_NAME        "Capture"

#define USBCAM_WIDTH        320
#define USBCAM_HEIGHT       240
#define USBCAM_DEF_PIX_FMT  V4L2_PIX_FMT_YUYV
#define YUV_PIXEL_FMT       PVRSRV_PIXEL_FORMAT_FOURCC_ORG_YUYV

#define CLEAR(x)            memset(&(x), 0, sizeof(x))

extern int gQuit;
extern char *cap_dev;
static buf_info_t buff_info[NUM_BUFFERS];
static int fd = -1;

int frame_init(bc_buf_params_t *p)
{
    int i, j, index;
    struct v4l2_requestbuffers reqbuf;
    struct v4l2_buffer buf;
    struct v4l2_capability capability;
    struct v4l2_input input;
    struct v4l2_format fmt;
    int retry = 0;

    if(!cap_dev)
        cap_dev = CAPTURE_DEVICE;

    /* Open the capture device */
    while (1) {
        if (gQuit == TRUE || retry >= 10)
            return -1;

        fd = open(cap_dev, O_RDWR|O_NONBLOCK);
        if (fd  <= 0) {
            printf("Cannot open capture device\n");
            printf("Please check the webcam... retry in 3 seconds...\n");

            sleep(3);
            retry++;
            continue;
        }
        break;
    }

    /* Get any active input */
    if (ioctl(fd, VIDIOC_G_INPUT, &index) < 0) {
        perror("VIDIOC_G_INPUT");
        goto ERROR;
    }

    /* Enumerate input to get the name of the input detected */
    CLEAR(input);
    input.index = index;
    if (ioctl(fd, VIDIOC_ENUMINPUT, &input) < 0) {
        perror("VIDIOC_ENUMINPUT");
        goto ERROR;
    }
	printf("INFO: current input is %s\n", input.name);
    index = 0;
    if (ioctl(fd, VIDIOC_S_INPUT, &index) < 0) {
        perror("VIDIOC_S_INPUT");
        goto ERROR;
    }

    CLEAR(input);
    input.index = index;
    if (ioctl(fd, VIDIOC_ENUMINPUT, &input) < 0) {
        perror("VIDIOC_ENUMINPUT");
        goto ERROR;
    }

    /* Check if the device is capable of streaming */
    if (ioctl(fd, VIDIOC_QUERYCAP, &capability) < 0) {
        perror("VIDIOC_QUERYCAP");
        goto ERROR;
    }
    if (!capability.capabilities & V4L2_CAP_STREAMING) {
        printf("%s: Not capable of streaming\n", CAPTURE_NAME);
        goto ERROR;
    }

    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(fd, VIDIOC_G_FMT, &fmt) < 0) {
        perror("VIDIOC_G_FMT");
        goto ERROR;
    }

    fmt.fmt.pix.width = USBCAM_WIDTH;
    fmt.fmt.pix.height = USBCAM_HEIGHT;
    fmt.fmt.pix.pixelformat = USBCAM_DEF_PIX_FMT;

    if (ioctl(fd, VIDIOC_S_FMT, &fmt) < 0) {
        perror("VIDIOC_S_FMT");
        goto ERROR;
    }

    if (ioctl(fd, VIDIOC_G_FMT, &fmt) < 0) {
        perror("VIDIOC_G_FMT");
        goto ERROR;
    }

    if (fmt.fmt.pix.pixelformat != USBCAM_DEF_PIX_FMT) {
        printf("%s: Requested pixel format not supported\n", CAPTURE_NAME);
        goto ERROR;
    }

    /* Buffer allocation */
    CLEAR(reqbuf);
    reqbuf.count = NUM_BUFFERS;
    reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    reqbuf.memory = V4L2_MEMORY_MMAP;
    if (ioctl(fd, VIDIOC_REQBUFS, &reqbuf) < 0) {
        perror("Cannot allocate memory");
        goto ERROR;
    }
    /*TODO check if NUM_BUFFERS got allocated*/

    for(i = 0; i < NUM_BUFFERS; i++)
        buff_info[i].start = NULL;

    CLEAR(buf);

    for (i = 0; i < NUM_BUFFERS; i++) {
        buf.type = reqbuf.type;
        buf.index = i;
        buf.memory = reqbuf.memory;
        if (ioctl(fd, VIDIOC_QUERYBUF, &buf) < 0) {
            perror("VIDIOC_QUERYCAP");
            goto ERROR1;
        }

        buff_info[i].length = buf.length;
        buff_info[i].index = i;
        buff_info[i].start = mmap(NULL, buf.length,
                PROT_READ | PROT_WRITE, MAP_SHARED,
                fd, buf.m.offset);

        if (buff_info[i].start == MAP_FAILED) {
            printf("Cannot mmap = %d buffer\n", i);
            goto ERROR1;
        }

        if (ioctl(fd, VIDIOC_QBUF, &buf) < 0) {
            perror("VIDIOC_QBUF");
            goto ERROR1;
        }
    }

    /* Start Streaming on capture device */
    index = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(fd, VIDIOC_STREAMON, &index) < 0) {
        perror("VIDIOC_STREAMON");
        goto ERROR1;
    }

    if (p) {
        p->count = 1;
        p->width = USBCAM_WIDTH;
        p->height = USBCAM_HEIGHT;
        p->pixel_fmt = YUV_PIXEL_FMT;
        p->type = BC_MEMORY_MMAP;      
    }
    return 0;

ERROR1:
    for (j = 0; j < NUM_BUFFERS; j++)
        if (buff_info[j].start)
            munmap(buff_info[j].start, buff_info[j].length);
ERROR:
    close(fd);

    return -1;
}

/*****************************************************************************
 * return cap buffer virtual address
 *      NULL - buffer unavailable
 *        -1 - DQBUF failed
 ****************************************************************************/
char *frame_get(bc_buf_ptr_t *p)
{
    struct v4l2_buffer buf;

    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    /* Dequeue capture buffer */
    if (ioctl(fd, VIDIOC_DQBUF, &buf) < 0) {
        if (errno == EAGAIN)
            return NULL;
        perror("VIDIOC_DQBUF");
        return (char *)-1;
    }

    if (p)
        p->pa = 0;

    return buff_info[buf.index].start;
}

int frame_restore(char *p)
{
    struct v4l2_buffer buf;
    int ii;

    /*dont bother with NULL pointer*/
    if (!p)
        return 0;

    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    for (ii = 0; ii < NUM_BUFFERS; ii++)
        if (buff_info[ii].start == p)
            break;

    /*buffer not found*/
    if (ii == NUM_BUFFERS)
        return -1;

    buf.index = ii;

    /* Queue capture buffer */
    if (ioctl(fd, VIDIOC_QBUF, &buf) < 0) {
        perror("VIDIOC_DQBUF");
        return -1;
    }
    return 0;
}

void frame_cleanup(void)
{
    int i, a;

    a = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(fd, VIDIOC_STREAMOFF, &a))
        perror("VIDIOC_STREAMOFF");

    for (i = 0; i < NUM_BUFFERS; i++) {
        munmap(buff_info[i].start, buff_info[i].length);
        buff_info[i].start = NULL;
    }
    close(fd);
}
