/*****************************************************************************
 * main.c
 *
 * bc-cat module unit test
 *
 * Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *   
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the
 *   distribution.
 *   
 *   Neither the name of Texas Instruments Incorporated nor the names of
 *   its contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/fb.h>

#ifdef X11
#include "X11/Xlib.h"
#include "X11/Xutil.h"
#endif

#include <GLES/gl.h>
#include <GLES/egl.h>
#include <GLES/glext.h>
#include <bc_cat.h>

#define MAX_BUFFERS     1
#define TRUE            1
#define FALSE           0
#define V_LEN           1.8

typedef struct tex_buffer_info {
    int n;
    int w;
    int h;
    int fmt;
} tex_buffer_info_t;

int frame_init(bc_buf_params_t *p);
char *frame_get(bc_buf_ptr_t *p);
int frame_restore(char *p);
void frame_cleanup(void);


int gQuit = FALSE;
int profiling = FALSE;
char *cap_dev = NULL;

#ifdef X11
Display* x11Display  = 0;
Window   x11Window   = 0;
Colormap x11Colormap = 0;
#endif

EGLDisplay dpy;
EGLSurface surface = EGL_NO_SURFACE;
EGLContext context = EGL_NO_CONTEXT;

PFNGLTEXBINDSTREAMIMGPROC glTexBindStreamIMG = NULL;


void signalHandler(int signum) { (void)signum; gQuit=TRUE; }

int get_disp_resolution(int *w, int *h)
{
    int fb_fd, ret = -1;
    struct fb_var_screeninfo vinfo;

    if ((fb_fd = open("/dev/fb0", O_RDONLY)) < 0) {
        printf("failed to open fb0 device\n");
        return ret;
    }

    if (ioctl(fb_fd, FBIOGET_VSCREENINFO, &vinfo) < 0) {
        perror("FBIOGET_VSCREENINFO");
        goto exit;
    }

    *w = vinfo.xres;
    *h = vinfo.yres;

    if (*w && *h)
        ret = 0;

exit:
    close(fb_fd);
    return ret;
}

#ifdef X11
int initX11(int *w, int *h)
{
    long                 screen = 0;
    XVisualInfo          visual;
    Window               rootWin;
    XSetWindowAttributes wa;
    unsigned int         mask;
    int                  depth;
    int                  width;
    int                  height;

    if (!(x11Display = XOpenDisplay(":0"))) {
        printf("Error: Unable to open X display\n");
        return -1;
    }
    screen = XDefaultScreen(x11Display);

    width = DisplayWidth(x11Display, screen);
    height = DisplayHeight(x11Display, screen);
    rootWin = RootWindow(x11Display, screen);
    depth = DefaultDepth(x11Display, screen);

    if (!XMatchVisualInfo(x11Display, screen, depth, TrueColor, &visual)) {
        printf("Error: Unable to acquire visual\n");
        XCloseDisplay(x11Display);
        return -1;
    }
    x11Colormap = XCreateColormap(x11Display, rootWin,
                                  visual.visual, AllocNone);
    wa.colormap = x11Colormap;

    wa.event_mask = StructureNotifyMask | ExposureMask;
    mask = CWBackPixel | CWBorderPixel | CWEventMask | CWColormap;

    x11Window = XCreateWindow(x11Display, rootWin,
                              0, 0, width/4*3, height/4*3,
                              0, CopyFromParent, InputOutput,
                              CopyFromParent, mask, &wa);
    XMapWindow(x11Display, x11Window);
    XFlush(x11Display);

    if (w) *w = width;
    if (h) *h = height;

    return 0;
}

void deInitX11(void)
{
    if (x11Window)
        XDestroyWindow(x11Display, x11Window);
    if (x11Colormap)
        XFreeColormap(x11Display, x11Colormap);
    if (x11Display)
        XCloseDisplay(x11Display);
}

int doX11Events(void)
{
    int ii, w, h;
    float r;
    XEvent e;
    Atom wdw;

    wdw = XInternAtom(x11Display, "WM_DELETE_WINDOW", False); 
    XSetWMProtocols(x11Display, x11Window, &wdw, 1);

    for (ii = 0; ii < XPending(x11Display); ii++) {
        XNextEvent(x11Display, &e);
 
        switch (e.type) {
        case ClientMessage:
            /* WM_DELETE_WINDOW - terminate */
            if ((Atom)e.xclient.data.l[0] == wdw)
                return 1;
            break; 
        case ConfigureNotify:
            w = e.xconfigure.width;
            h = e.xconfigure.height;
            r = (float)w / h;

            glViewport(0, 0, w, h);
            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();

            if (w > h)
                glOrthof(-V_LEN * r, V_LEN * r, -V_LEN, V_LEN,
                         -V_LEN * 2, V_LEN * 2);
            else
                glOrthof(-V_LEN, V_LEN, -V_LEN / r, V_LEN / r,
                         -V_LEN * 2, V_LEN * 2);

            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();

            break;
        case Expose:
            /*must handle this, otherwise, background does not redraw*/
            break;
        default:
            break;
        }
    }
    return 0;
}
#endif

int initTexExt(tex_buffer_info_t *binfo) 
{
    PFNGLGETTEXSTREAMDEVICEATTRIBUTEIVIMGPROC glGetTexAttrIMG = NULL;
    PFNGLGETTEXSTREAMDEVICENAMEIMGPROC glGetTexDeviceIMG = NULL;
    const GLubyte *glext;
    const GLubyte *dev_name;

    if (!binfo)
        return -5;

    if (!(glext = glGetString(GL_EXTENSIONS)))
        return -1;

    if (!strstr((char *)glext, "GL_IMG_texture_stream"))
        return -2;

    glTexBindStreamIMG =
        (PFNGLTEXBINDSTREAMIMGPROC)eglGetProcAddress("glTexBindStreamIMG");
    glGetTexAttrIMG = (PFNGLGETTEXSTREAMDEVICEATTRIBUTEIVIMGPROC)
        eglGetProcAddress("glGetTexStreamDeviceAttributeivIMG");
    glGetTexDeviceIMG = (PFNGLGETTEXSTREAMDEVICENAMEIMGPROC)
        eglGetProcAddress("glGetTexStreamDeviceNameIMG");

    if (!glTexBindStreamIMG || !glGetTexAttrIMG || !glGetTexDeviceIMG)
        return -3;

    dev_name = glGetTexDeviceIMG(0);
    if (!dev_name)
        return -4;

    glGetTexAttrIMG(0, GL_TEXTURE_STREAM_DEVICE_NUM_BUFFERS_IMG, &binfo->n);
    glGetTexAttrIMG(0, GL_TEXTURE_STREAM_DEVICE_WIDTH_IMG, &binfo->w);
    glGetTexAttrIMG(0, GL_TEXTURE_STREAM_DEVICE_HEIGHT_IMG, &binfo->h);
    glGetTexAttrIMG(0, GL_TEXTURE_STREAM_DEVICE_FORMAT_IMG, &binfo->fmt);

    printf("\ndevice: %s num: %d, width: %d, height: %d, format: 0x%x\n",
        dev_name, binfo->n, binfo->w, binfo->h, binfo->fmt);

    return 0;
}

void print_err(char *name)
{
    char *err_str[] = {
          "EGL_SUCCESS",
          "EGL_NOT_INITIALIZED",
          "EGL_BAD_ACCESS",
          "EGL_BAD_ALLOC",
          "EGL_BAD_ATTRIBUTE",    
          "EGL_BAD_CONFIG",
          "EGL_BAD_CONTEXT",   
          "EGL_BAD_CURRENT_SURFACE",
          "EGL_BAD_DISPLAY",
          "EGL_BAD_MATCH",
          "EGL_BAD_NATIVE_PIXMAP",
          "EGL_BAD_NATIVE_WINDOW",
          "EGL_BAD_PARAMETER",
          "EGL_BAD_SURFACE" };

    EGLint ecode = eglGetError();

    printf("'%s': egl error '%s' (0x%x)\n",
           name, err_str[ecode-EGL_SUCCESS], ecode);
}

void deInitEGL(void)
{
    eglMakeCurrent(dpy, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
    if (context != EGL_NO_CONTEXT)
        eglDestroyContext(dpy, context);
    if (surface != EGL_NO_SURFACE)
        eglDestroySurface(dpy, surface);
    eglTerminate(dpy);
}

int initEGL(void)
{
    EGLint            disp_w, disp_h;
    NativeDisplayType disp_type;
    NativeWindowType  window;
    EGLConfig         cfgs[2];
    EGLint            n_cfgs;
    EGLint            egl_attr[] = {
                         EGL_BUFFER_SIZE, EGL_DONT_CARE,
                         EGL_RED_SIZE,    5,
                         EGL_GREEN_SIZE,  6,
                         EGL_BLUE_SIZE,   5,
                         EGL_DEPTH_SIZE,  8,
                         EGL_NONE };

#ifdef X11
    if (initX11(&disp_w, &disp_h))
        return -1;

    disp_type = (NativeDisplayType)x11Display;
    window = (NativeWindowType)x11Window;
#else
    if (get_disp_resolution(&disp_w, &disp_h)) {
        printf("ERROR: get display resolution failed\n");
        return -1;
    }

    disp_type = EGL_DEFAULT_DISPLAY;
    window  = 0;
#endif

    dpy = eglGetDisplay(disp_type);

    if (eglInitialize(dpy, NULL, NULL) != EGL_TRUE) {
        print_err("eglInitialize");
        return -1;
    }

    if (eglGetConfigs(dpy, cfgs, 2, &n_cfgs) != EGL_TRUE) {
        print_err("eglGetConfigs");
        goto cleanup;
    }
    
    if (eglChooseConfig(dpy, egl_attr, cfgs, 2, &n_cfgs) != EGL_TRUE) {
        print_err("eglChooseConfig");
        goto cleanup;
    }

    surface = eglCreateWindowSurface(dpy, cfgs[0], window, NULL);
    if (surface == EGL_NO_SURFACE) {
        print_err("eglCreateWindowSurface");
        goto cleanup;
    }

    context = eglCreateContext(dpy, cfgs[0], EGL_NO_CONTEXT, NULL);
    if (context == EGL_NO_CONTEXT) {
        print_err("eglCreateContext");
        goto cleanup;
    }

    if (eglMakeCurrent(dpy, surface, surface, context) != EGL_TRUE) {
        print_err("eglMakeCurrent");
        goto cleanup;
    }
  
    /* 0 - do not sync with video frame */
    if (profiling == TRUE) {
        if (eglSwapInterval(dpy, 0) != EGL_TRUE) {
            print_err("eglSwapInterval");
            goto cleanup;
        }
    }
  
    glShadeModel(GL_FLAT);
    glTexParameterf(GL_TEXTURE_STREAM_IMG, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_STREAM_IMG, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    glFrontFace(GL_CW);
    glCullFace(GL_FRONT);
    glEnable(GL_CULL_FACE);

    glEnable(GL_NORMALIZE);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glOrthof(-V_LEN * disp_w / disp_h, V_LEN * disp_w / disp_h,
             -V_LEN, V_LEN, -V_LEN * 2, V_LEN * 2);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    return 0;

cleanup:
    deInitEGL();
#ifdef X11
    deInitX11();
#endif
    return -1;
}

static GLfloat Vertices[16][3] =
{   // x     y     z
    {-1.0, -1.0,  1.0}, // 1  left    First Strip
    {-1.0,  1.0,  1.0}, // 3
    {-1.0, -1.0, -1.0}, // 0
    {-1.0,  1.0, -1.0}, // 2
    { 1.0, -1.0, -1.0}, // 4  back
    { 1.0,  1.0, -1.0}, // 6
    { 1.0, -1.0,  1.0}, // 5  right
    { 1.0,  1.0,  1.0}, // 7

    { 1.0,  1.0, -1.0}, // 6  top     Second Strip
    {-1.0,  1.0, -1.0}, // 2
    { 1.0,  1.0,  1.0}, // 7
    {-1.0,  1.0,  1.0}, // 3
    { 1.0, -1.0,  1.0}, // 5  front
    {-1.0, -1.0,  1.0}, // 1
    { 1.0, -1.0, -1.0}, // 4  bottom
    {-1.0, -1.0, -1.0}  // 0
};

static GLfloat NormalsPerVertex[16][3] =    // One normal per vertex.
{   // x     y     z
    {-0.5, -0.5,  0.5}, // 1  left          First Strip
    {-0.5,  0.5,  0.5}, // 3
    {-0.5, -0.5, -0.5}, // 0
    {-0.5,  0.5, -0.5}, // 2
    { 0.5, -0.5, -0.5}, // 4  back
    { 0.5,  0.5, -0.5}, // 6
    { 0.5, -0.5,  0.5}, // 5  right
    { 0.5,  0.5,  0.5}, // 7

    { 0.5,  0.5, -0.5}, // 6  top           Second Strip
    {-0.5,  0.5, -0.5}, // 2
    { 0.5,  0.5,  0.5}, // 7
    {-0.5,  0.5,  0.5}, // 3
    { 0.5, -0.5,  0.5}, // 5  front
    {-0.5, -0.5,  0.5}, // 1
    { 0.5, -0.5, -0.5}, // 4  bottom
    {-0.5, -0.5, -0.5}  // 0
};

static GLfloat TexCoords[16][2] =
{   // x   y
    {0.0, 1.0}, // 1  left                  First Strip
    {0.0, 0.0}, // 3
    {1.0, 1.0}, // 0
    {1.0, 0.0}, // 2
    {0.0, 1.0}, // 4  back
    {0.0, 0.0}, // 6
    {1.0, 1.0}, // 5  right
    {1.0, 0.0}, // 7

    {0.0, 1.0}, // 1  top                   Second Strip
    {1.0, 1.0}, // 3
    {0.0, 0.0}, // 0
    {1.0, 0.0}, // 2
    {0.0, 1.0}, // 4  front
    {1.0, 1.0}, // 6
    {0.0, 0.0}, // 5  bottom
    {1.0, 0.0}  // 7
};

void drawCube(int bufferindex)
{
    static GLfloat m_fAngleX = 0.0f;
    static GLfloat m_fAngleY = 0.0f;

    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_TEXTURE_STREAM_IMG);

    glPushMatrix();

    // Rotate the cube model
    glRotatef(m_fAngleX, 1.0f, 0.0f, 0.0f);
    glRotatef(m_fAngleY, 0.0f, 1.0f, 0.0f);

    glTexBindStreamIMG(0, bufferindex);
/*    glBindTexture(GL_TEXTURE_STREAM_IMG, bufferindex);*/

    // Enable 3 types of data
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    // Set pointers to the arrays
    glVertexPointer(3, GL_FLOAT, 0, Vertices);
    glNormalPointer(GL_FLOAT, 0, NormalsPerVertex);
    glTexCoordPointer(2, GL_FLOAT, 0, TexCoords);

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 8);
    glDrawArrays(GL_TRIANGLE_STRIP, 8, 8);

    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);

    glPopMatrix();


    glDisable(GL_TEXTURE_STREAM_IMG);
    eglSwapBuffers(dpy, surface);

    m_fAngleX += 0.25f;
    m_fAngleY += 0.25f;
}

void usage(char *arg)
{
    printf("Usage:\n"
           "    %s [-c <dev> | -p | -w <width> | -t <height> | -h]\n"
           "\t-c - capture device [default: /dev/video0]\n"
           "\t-p - enable profiling\n"
           "\t-w - width of texture buffer in pixel, multiple of 32\n"
           "\t-t - height of texture buffer in pixel\n"
           "\t-h - print this message\n\n", arg);
}

int main(int argc, char *argv[])
{
    int bcfd = -1; 
    BCIO_package ioctl_var;
    bc_buf_params_t buf_param;
    bc_buf_ptr_t buf_pa;
    tex_buffer_info_t buf_info;

    unsigned long buf_paddr[MAX_BUFFERS];
    char *buf_vaddr[MAX_BUFFERS] = { MAP_FAILED };
    char *frame = NULL;
    int buf_size = 0;
    int c, idx, ret = -1;
    char opts[] = "c:pw:t:h";

    int   ii;
    int   frame_w, frame_h;
    int   min_w = 0, min_h = 0;;
    int   cp_offset = 0;

    struct timeval tvp, tv, tv0 = {0,0};
    unsigned long tdiff = 0;
    unsigned long fcount = 0;

    for (;;) {
        c = getopt_long(argc, argv, opts, (void *)NULL, &idx);

        if (-1 == c)
            break;

        switch (c) {
            case 0:
                break;
            case 'c':
                cap_dev = optarg;
                printf("INFO: capture device is %s\n", cap_dev);
                break;
            case 'p':
                profiling = TRUE;
                printf("INFO: profiling enabled\n");
                break;
            case 'w':
                min_w = atoi(optarg);
                break;
            case 't':
                min_h = atoi(optarg);
                break;
            default:
                usage(argv[0]);
                return 0;
        }
    }

    signal(SIGINT, signalHandler);

    if (frame_init(&buf_param))
        return -1;

    if ((bcfd = open("/dev/bc_cat", O_RDWR|O_NDELAY)) == -1) {
        printf("ERROR: open '/dev/bc_cat' failed\n");
        goto err_ret;
    }

    frame_w = buf_param.width;
    frame_h = buf_param.height;

    if (min_w > 0 && !(min_w % 32))
        buf_param.width = min_w;

    if (min_h > 0)
        buf_param.height = min_h;
 
    if (ioctl(bcfd, BCIOREQ_BUFFERS, &buf_param) != 0) {
        printf("ERROR: BCIOREQ_BUFFERS failed\n");
        goto err_ret;
    }

    if (ioctl(bcfd, BCIOGET_BUFFERCOUNT, &ioctl_var) != 0) {
        goto err_ret;
    }

    if (ioctl_var.output == 0) {
        printf("ERROR: no texture buffer available\n");
        goto err_ret;
    }

    /* for BC_MEMORY_USERPTR, BCIOSET_BUFFERPHYADDR must be called
     * before init IMG_extensions in initTexExt()*/
    if (buf_param.type == BC_MEMORY_USERPTR) {
        for (idx = 0; idx < buf_param.count; idx++) {
            while ((frame = frame_get(&buf_pa)) == NULL) { }

            if (frame == (char *)-1)
                goto err_ret;

            if (ioctl(bcfd, BCIOSET_BUFFERPHYADDR, &buf_pa) != 0) {
                frame_restore(frame);
                printf("ERROR: BCIOSET_BUFFERADDR[%d]: failed (0x%lx)\n",
                       buf_pa.index, buf_pa.pa);
                goto err_ret;
            }

            if (frame_restore(frame))
                goto err_ret;
        }
    }

    if (initEGL()) {
        printf("ERROR: init EGL failed\n");
        goto err_ret;
    }
 
    if ((ret = initTexExt(&buf_info)) < 0) {
        printf("ERROR: initTexExt() failed [%d]\n", ret);
        goto err_ret;
    }

    /*FIXME calc stride instead of 2*/
    buf_size = buf_info.w * buf_info.h * 2;
    min_w    = buf_info.w < frame_w ? buf_info.w : frame_w;
    min_h    = buf_info.h < frame_h ? buf_info.h : frame_h;

    if (buf_info.h > frame_h)
        cp_offset = (buf_info.h - frame_h) * buf_info.w;

    if (buf_info.w > frame_w)
        cp_offset += buf_info.w - frame_w;

    if (buf_param.type == BC_MEMORY_MMAP) {
        for (idx = 0; idx < MAX_BUFFERS; idx++) {
            ioctl_var.input = idx;

            if (ioctl(bcfd, BCIOGET_BUFFERPHYADDR, &ioctl_var) != 0) {
                printf("ERROR: BCIOGET_BUFFERADDR failed\n");
                goto err_ret;
            }

            buf_paddr[idx] = ioctl_var.output;
            buf_vaddr[idx] = (char *)mmap(NULL, buf_size,
                              PROT_READ | PROT_WRITE, MAP_SHARED,
                              bcfd, buf_paddr[idx]);

            if (buf_vaddr[idx] == MAP_FAILED) {
                printf("ERROR: mmap failed\n");
                goto err_ret;
            }
        }
    }

    ret = 0;
    idx = 0;

    if (profiling == TRUE) {
        gettimeofday(&tvp, NULL);
        tv0 = tvp;
    }

    while (!gQuit) {
        frame = frame_get(&buf_pa);

        if (frame == (char *) -1)
            break;

        if (frame) {
            if (buf_param.type == BC_MEMORY_MMAP) {
                for (ii = 0; ii < min_h; ii++)
                      /*FIXME calc stride instead of 2*/
                    memcpy(buf_vaddr[idx] + buf_info.w * 2 * ii + cp_offset,
                           frame + frame_w * 2 * ii, min_w * 2);
            }
            else    /*buf_param.type == BC_MEMORY_USERPTR*/
                idx = buf_pa.index;
        }

        drawCube(idx);

        if (frame_restore(frame))
            break;

#ifdef X11
        if (doX11Events())
            gQuit = TRUE;
#endif

        if (profiling == FALSE)
            continue;

        gettimeofday(&tv, NULL);

        fcount++;

        if (!(fcount % 60)) {
            tdiff = (unsigned long)(tv.tv_sec*1000 + tv.tv_usec/1000 -
                                tvp.tv_sec*1000 - tvp.tv_usec/1000);
            if (tdiff < 1800)   /*print fps every 2 sec*/
                continue;

            fprintf(stderr, "\rAvg FPS: %ld",
                    fcount / (tv.tv_sec - tv0.tv_sec));
            tvp = tv;
        }
    }
    printf("\n");

err_ret:
    if (buf_param.type == BC_MEMORY_MMAP) {
        for (idx = 0; idx < MAX_BUFFERS; idx++) {
            if (buf_vaddr[idx] != MAP_FAILED)
                munmap(buf_vaddr[idx], buf_size);
        }
    }
    if (bcfd > -1)
        close(bcfd);

    deInitEGL();
#ifdef X11
    deInitX11();
#endif
    frame_cleanup();

    printf("done\n");
    return ret;
}

